"""
:mod:`transformer.plugins.contracts` -- Contracts for Plugin Authors
====================================================================

This module defines the various :ref:`Transformer contracts <contracts>`
and their helper functions.
"""
import enum
from collections import defaultdict
from typing import NewType, Iterable, TypeVar, List, DefaultDict


class Contract(enum.Flag):
    """
    Enumeration of all supported :ref:`plugin contracts <contracts>`.
    Each specific contract defines a way for plugins to be used in Transformer.

    Any Python function may become a Transformer plugin by announcing that
    it implements at least one contract, using the :func:`@plugin <plugin>`
    decorator.
    """

    OnTask = enum.auto()  #: The :term:`OnTask` contract.
    OnScenario = enum.auto()  #: The :term:`OnScenario` contract.
    OnPythonProgram = enum.auto()  #: The :term:`OnPythonProgram` contract.

    # Historically Transformer has only one plugin contract, which transformed a
    # sequence of Task objects into another such sequence. Operating on a full list
    # of tasks (instead of task by task) offered more leeway: a plugin could e.g.
    # add a new task, or change only the first task.
    # However this OnTaskSequence model is too constraining for some use-cases,
    # e.g. when a plugin needs to inject code in the global scope, and having to
    # deal with a full, immutable list of tasks in plugins that independently
    # operate on each task implies a lot of verbosity and redundancy.
    # For these reasons, other plugin contracts were created to offer a more
    # varied choice for plugin implementers.
    # See https://github.com/zalando-incubator/Transformer/issues/10.
    OnTaskSequence = enum.auto()  #: Deprecated.


Plugin = NewType("Plugin", callable)


class InvalidContractError(ValueError):
    """
    Raised for plugin functions associated with invalid contracts.

    What an "invalid contract" represents is not strictly specified,
    but this includes at least objects that are not members of the Contract
    enumeration.
    """


class InvalidPluginError(ValueError):
    """
    Raised when trying to use as plugin a function that has not been marked
    as such.
    """


def plugin(c: Contract):
    """
    Documented in dev.rst.
    """
    if not isinstance(c, Contract):
        suggestions = (f"@plugin(Contract.{x.name})" for x in Contract)
        raise InvalidContractError(
            f"{c!r} is not a {Contract.__qualname__}. "
            f"Did you mean {', '.join(suggestions)}?"
        )

    def _decorate(f: callable) -> callable:
        f._transformer_plugin_contract = c
        return f

    return _decorate


def contract(f: Plugin) -> Contract:
    """
    Returns the contract associated to a plugin function.

    :raise InvalidPluginError: if f is not a plugin.
    """
    try:
        return getattr(f, "_transformer_plugin_contract")
    except AttributeError:
        raise InvalidPluginError(f) from None


_T = TypeVar("_T")


def apply(plugins: Iterable[Plugin], init: _T) -> _T:
    """
    Applies each plugin to init in order, and returns the result.

    This just wraps a very simple but common operation.
    """
    for p in plugins:
        init = p(init)
    return init


_BASE_CONTRACTS = (
    Contract.OnTask,
    Contract.OnTaskSequence,
    Contract.OnScenario,
    Contract.OnPythonProgram,
)


def group_by_contract(plugins: Iterable[Plugin]) -> DefaultDict[Contract, List[Plugin]]:
    """
    Groups plugins in lists according to their contracts.
    Each plugin is found in as many lists as it implements base contracts.
    Lists keep the order of the original plugins iterable.
    """
    res = defaultdict(list)
    for p in plugins:
        c = contract(p)
        for bc in _BASE_CONTRACTS:
            if c & bc:  # Contract is an enum.Flag: & computes the intersection.
                res[bc].append(p)
    return res



from transformer.plugins import plugin, Contract
from transformer.task import Task2
from transformer.request import CaseInsensitiveDict

COOKIE_KEY = "cookie"


@plugin(Contract.OnTask)
def plugin(task: Task2) -> Task2:
    """
    Removes Chrome-specific, RFC-non-compliant headers starting with `:`.
    Removes the cookie header as it is handled by Locust's HttpSession.
    """
    task.request.headers = CaseInsensitiveDict(
        {k: v for k, v in task.request.headers.items() if not k.startswith(":")}
    )
    if COOKIE_KEY in task.request.headers:
        task.request.headers.pop(COOKIE_KEY)

    return task



"""
:mod:`transformer.python` -- Python Syntax Tree
===============================================

Transformer's Python Syntax Tree framework allows you to create and manipulate
Python source code without bothering with irrelevant, style-related details.

It is the main API for writing :term:`OnPythonProgram` plugins.

A non-goal of this framework is *customization of style*: users should rely on
an external tool (such as `black`_) if they need style customization of their
generated locustfile.

.. _black: https://github.com/ambv/black
"""

import re
from types import MappingProxyType
from typing import (
    Sequence,
    Mapping,
    Any,
    List,
    Type,
    Set,
    Optional,
    Tuple,
    cast,
    Iterable,
    Callable,
    TypeVar,
    ClassVar,
)

from dataclasses import dataclass

IMMUTABLE_EMPTY_DICT = MappingProxyType({})


@dataclass
class Line:
    """
    A line of text and its associated indentation level.

    This class allows not to constantly copy strings to add a new indentation
    level at every scope of the syntax tree.

    .. attribute:: text

       :any:`str` -- Text contained by this line.

    .. attribute:: indent_level

       :any:`int` -- Indentation level of :attr:`text` in the line.
    """

    text: str
    indent_level: int = 0

    INDENT_UNIT: ClassVar[str] = " " * 4

    def __str__(self) -> str:
        """
        Textual representation of this line, with :attr:`text` indented
        according to :attr:`indent_level`.
        """
        return f"{self.INDENT_UNIT * self.indent_level}{self.text}"

    def clone(self) -> "Line":
        """
        Creates an exact but disconnected copy of self.
        Useful in tests.
        """
        return type(self)(text=self.text, indent_level=self.indent_level)


def _resplit(parts: Iterable[str]) -> List[str]:
    """
    Given a list of strings, returns a list of lines, by splitting each string
    into multiple lines where it contains newlines.

    >>> _resplit([])
    []
    >>> _resplit(['a', 'b'])
    ['a', 'b']
    >>> _resplit(['a', 'b\\nc\\nd'])
    ['a', 'b', 'c', 'd']
    """
    return [line for part in parts for line in part.splitlines()]


class Statement:
    """
    Python distinguishes between statements and expressions: basically,
    statements cannot be assigned to a variable, whereas expressions can.

    For our purpose, another distinction is important: statements may span over
    multiple lines (and not just for style), whereas all expressions can be
    expressed in a single line.

    This class serves as abstract base for all implementors of :meth:`lines` and
    handles comment processing for them.
    """

    def __init__(self, comments: Sequence[str] = ()) -> None:
        """
        :param comments: Comment lines attached to this statement.
        """
        self._comments = _resplit(comments)

    @property
    def comments(self) -> List[str]:
        """
        Comment lines attached to this statement.

        This is a :class:`property` to ensure that modifications of this list
        preserve the invariant "one element = one line".
        """
        self._comments = _resplit(self._comments)
        return self._comments

    @comments.setter
    def comments(self, value: List[str]):
        self._comments = value

    def lines(self, indent_level: int = 0, comments: bool = True) -> List[Line]:
        """
        All Line objects necessary to represent this Statement, along with the
        appropriate indentation level.

        :param indent_level: How much indentation to apply to the least indented
            line of this statement.
        :param comments: Whether existing comments attached to *self* should be
            included in the result.
        """
        raise NotImplementedError

    def comment_lines(self, indent_level: int) -> List[Line]:
        """
        Converts self.comments from str to Line with ``#`` prefixes.
        """
        return [Line(f"# {s}", indent_level) for s in self.comments]

    def attach_comment(self, line: Line) -> List[Line]:
        """
        Attach a comment to *line*: inline if *self.comments* is just one line,
        on dedicated new lines above otherwise.
        """
        comments = self.comments
        if not comments:
            return [line]
        if len(comments) == 1:
            line.text += f"  # {comments[0]}"
            return [line]
        lines = self.comment_lines(line.indent_level)
        lines.append(line)
        return lines

    def __eq__(self, o: object) -> bool:
        return (
            isinstance(o, self.__class__)
            and self.comments == cast(__class__, o).comments
        )


# Handy alias for type signatures.
Program = Sequence[Statement]


class OpaqueBlock(Statement):
    """
    A block of code already represented as a string.
    This helps moving existing code (e.g. in plugins) from our ad-hoc
    "blocks of code" framework to the syntax tree framework defined in this
    module.
    It also allows to express Python constructs that would otherwise not yet be
    representable with this AST framework.
    """

    PREFIX_RX = re.compile(r"\s+")
    TAB_SIZE = 8

    def __init__(self, block: str, comments: Sequence[str] = ()) -> None:
        """
        :param block: String representing a block of Python code.
        """
        super().__init__(comments)
        if not block.strip():
            raise ValueError(f"OpaqueBlock can't be empty but got {block!r}")
        self.block = block

    def lines(self, indent_level: int = 0, comments: bool = True) -> List[Line]:
        raw_lines = [line.expandtabs(self.TAB_SIZE) for line in self.block.splitlines()]

        first_nonempty_line = next(i for i, l in enumerate(raw_lines) if l.strip())
        after_last_nonempty_line = next(
            len(raw_lines) - i for i, l in enumerate(reversed(raw_lines)) if l.strip()
        )
        raw_lines = raw_lines[first_nonempty_line:after_last_nonempty_line]

        indents = [self.PREFIX_RX.match(line) for line in raw_lines]
        shortest_indent = min(len(p.group()) if p else 0 for p in indents)
        block_lines = [Line(line[shortest_indent:], indent_level) for line in raw_lines]
        if comments:
            return [*self.comment_lines(indent_level), *block_lines]
        return block_lines

    def __repr__(self) -> str:
        return "{}({!r}, comments={!r})".format(
            self.__class__.__qualname__, self.block, self.comments
        )

    def __eq__(self, o: object) -> bool:
        return super().__eq__(o) and self.block == cast(__class__, o).block


class Function(Statement):
    """
    A function definition (``def ...``).
    """

    def __init__(
        self,
        name: str,
        params: Sequence[str],
        statements: Sequence[Statement],
        comments: Sequence[str] = (),
    ) -> None:
        """
        :param name: Name of this function.
        :param params: Names of each parameter of this function.
        :param statements: Body of this function.
        """
        super().__init__(comments)
        self.name = name
        self.params = list(params)
        self.statements = list(statements)

    def lines(self, indent_level: int = 0, comments: bool = True) -> List[Line]:
        param_list = ", ".join(self.params)
        body_lines = [
            line
            for stmt in self.statements
            for line in stmt.lines(indent_level + 1, comments)
        ] or [Line("pass", indent_level + 1)]
        top = Line(f"def {self.name}({param_list}):", indent_level)
        if comments:
            return [*self.attach_comment(top), *body_lines]
        return [top, *body_lines]

    def __repr__(self) -> str:
        return "{}(name={!r}, params={!r}, statements={!r}, comments={!r})".format(
            self.__class__.__qualname__,
            self.name,
            self.params,
            self.statements,
            self.comments,
        )

    def __eq__(self, o: object) -> bool:
        return (
            super().__eq__(o)
            and self.name == cast(__class__, o).name
            and self.params == cast(__class__, o).params
            and self.statements == cast(__class__, o).statements
        )


class Decoration(Statement):
    """
    A function or class definition to which is applied a decorator
    (e.g. ``@task``).
    """

    def __init__(
        self, decorator: str, target: Statement, comments: Sequence[str] = ()
    ) -> None:
        """
        :param decorator: Name of the decorator applied to *target*.
        :param target: Function or class definition to which is applied
            *decorator*.
        """
        super().__init__(comments)
        self.decorator = decorator
        self.target = target

    def lines(self, indent_level: int = 0, comments: bool = True) -> List[Line]:
        top = Line(f"@{self.decorator}", indent_level)
        target_lines = self.target.lines(indent_level, comments)
        if comments:
            return [*self.attach_comment(top), *target_lines]
        return [top, *target_lines]

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, comments={!r})".format(
            self.__class__.__qualname__, self.decorator, self.target, self.comments
        )

    def __eq__(self, o: object) -> bool:
        return (
            super().__eq__(o)
            and self.decorator == cast(__class__, o).decorator
            and self.target == cast(__class__, o).target
        )


class Class(Statement):
    """
    A class definition.
    """

    def __init__(
        self,
        name: str,
        statements: Sequence[Statement],
        superclasses: Sequence[str] = (),
        comments: Sequence[str] = (),
    ) -> None:
        """
        :param name: Name of this class.
        :param statements: Fields of this class: methods, attributes, etc.
        :param superclasses: Names of each superclass of this class.
            In fact anything in the "function argument" format can be used here,
            like keyword-arguments (but in a string!).
        """
        super().__init__(comments)
        self.name = name
        self.statements = list(statements)
        self.superclasses = list(superclasses)

    def lines(self, indent_level: int = 0, comments: bool = True) -> List[Line]:
        superclasses = ""
        if self.superclasses:
            superclasses = "({})".format(", ".join(self.superclasses))

        body = [
            line
            for stmt in self.statements
            for line in stmt.lines(indent_level + 1, comments)
        ] or [Line("pass", indent_level + 1)]

        top = Line(f"class {self.name}{superclasses}:", indent_level)
        if comments:
            return [*self.attach_comment(top), *body]
        return [top, *body]

    def __repr__(self) -> str:
        return (
            "{}(name={!r}, statements={!r}, " "superclasses={!r}, comments={!r})"
        ).format(
            self.__class__.__qualname__,
            self.name,
            self.statements,
            self.superclasses,
            self.comments,
        )

    def __eq__(self, o: object) -> bool:
        return (
            super().__eq__(o)
            and self.name == cast(__class__, o).name
            and self.statements == cast(__class__, o).statements
            and self.superclasses == cast(__class__, o).superclasses
        )


class Expression:
    """
    See the documentation of :class:`Statement` for why Expression is a separate
    class.
    An expression is still a statement in Python (e.g. functions can be called
    anywhere), but this :class:`Expression` class is **not** a
    :class:`Statement` because we can't attach comments to arbitrary expressions
    (e.g. between braces).
    If you need to use an :class:`Expression` as a :class:`Statement`,
    see the :class:`Standalone` wrapper class.

    This class serves as abstract base for all our implementors of
    :meth:`__str__`.
    """

    def __str__(self) -> str:
        raise NotImplementedError

    def __eq__(self, o: object) -> bool:
        return isinstance(o, self.__class__)


class Standalone(Statement):
    """
    Wraps an :class:`Expression` so that it can be used as a :class:`Statement`.
    """

    def __init__(self, expr: Expression, comments: Sequence[str] = ()) -> None:
        """
        :param expr: The wrapped expression.
        """
        super().__init__(comments)
        self.expr = expr

    def lines(self, indent_level: int = 0, comments: bool = True) -> List[Line]:
        """
        An :class:`Expression` E used as a :class:`Statement` is serialized as
        the result of :samp:`str({E})` on its own :class:`Line`.
        """
        line = Line(str(self.expr), indent_level)
        if comments:
            return self.attach_comment(line)
        return [line]

    def __repr__(self) -> str:
        return "{}({!r}, comments={!r})".format(
            self.__class__.__qualname__, self.expr, self.comments
        )

    def __eq__(self, o: object) -> bool:
        return super().__eq__(o) and self.expr == cast(__class__, o).expr


def _all_subclasses_of(cls: Type) -> Set[Type]:
    """
    All subclasses of *cls*, including non-direct ones (child of child of ...).
    """
    direct_subclasses = set(cls.__subclasses__())
    return direct_subclasses.union(
        s for d in direct_subclasses for s in _all_subclasses_of(d)
    )


class Literal(Expression):
    """
    All literal Python expressions (integers, strings, lists, etc.).

    Everything will be serialized using :func:`repr`, except :class:`Expression`
    objects that could be contained in a composite value like ``list``:
    they will be serialized with :func:`str`, as is probably expected.
    Thus:

    >>> str(Literal([1, {"a": FString("-{x}")}]))
    "[1, {'a': f'-{x}'}]"

    instead of something like ``[1, {'a': FString('-{x}')}]``.

    .. seealso:: :class:`FString`
    """

    def __init__(self, value: Any) -> None:
        """
        :param value: The Python literal represented by this node.
        """
        super().__init__()
        self.value = value

    _REPR_BY_EXPR_CLS = None

    def __str__(self) -> str:
        # This is not pretty, but repr() doesn't accept a visitor we could use
        # to say "just this time, use that code to serialize Expression objects".
        if Literal._REPR_BY_EXPR_CLS is None:
            Literal._REPR_BY_EXPR_CLS = {
                c: c.__repr__ for c in _all_subclasses_of(Expression)
            }
        try:
            for k in Literal._REPR_BY_EXPR_CLS.keys():
                k.__repr__ = k.__str__
            return repr(self.value)
        finally:
            for k, _repr in Literal._REPR_BY_EXPR_CLS.items():
                k.__repr__ = _repr

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self.value!r})"

    def __eq__(self, o: object) -> bool:
        return super().__eq__(o) and self.value == cast(__class__, o).value


class FString(Literal):
    """
    f-strings_ are strings that capture values from their environment.

    They cannot be handled in :class:`Literal` because they are a "trick" of the
    Python parser: *before* the program runs, they lose their ``f`` prefix and
    their template is evaluated, so when :class:`Literal` is instantiated, they
    are only a normal string that tried to capture values from Transformer's
    context (instead of *the locustfile's* context).

    .. _f-strings: https://docs.python.org/3/whatsnew/3.6.html#whatsnew36-pep498
    """

    def __init__(self, s: str) -> None:
        """
        :param s: The template of this f-string,
            for example ``a{x}b`` for the f-string ``f"a{x}b"``.
        """
        if not isinstance(s, str):
            raise TypeError(
                f"expecting a format string, got {s.__class__.__qualname__}: {s!r}"
            )
        super().__init__(s)

    def __str__(self) -> str:
        return "f" + repr(str(self.value))


class Symbol(Expression):
    """
    The name of something (variable, function, etc.).
    Avoids any kind of string quoting and escaping that would happen with
    :class:`Literal`.

    >>> str(Literal("x"))
    "'x'"
    >>> str(Symbol("x"))
    'x'

    The provided argument's type is explicitly checked and a :class:`TypeError`
    may be raised to avoid confusion when a user expects e.g. ``Symbol(True)``
    to work like ``Symbol("True")``.
    """

    def __init__(self, name: str) -> None:
        """
        :param name: Textual representation of this symbol.
            Will be forwarded without modification to the locustfile.
        """
        super().__init__()
        if not isinstance(name, str):
            raise TypeError(
                f"expected symbol name, got {name.__class__.__qualname__}: {name!r}"
            )
        self.name = name

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}({self.name!r})"

    def __eq__(self, o: object) -> bool:
        return super().__eq__(o) and self.name == cast(__class__, o).name


class FunctionCall(Expression):
    """
    The invocation of a function or method.
    """

    def __init__(
        self,
        name: str,
        positional_args: Sequence[Expression] = (),
        named_args: Mapping[str, Expression] = IMMUTABLE_EMPTY_DICT,
    ) -> None:
        """
        :param name: Name of the function that is called.
        :param positional_args: Positional arguments associated with this call,
            if any.
        :param named_args: Keyword-arguments associated with this call, if any.
        """
        super().__init__()
        self.name = name
        self.positional_args = list(positional_args)
        self.named_args = dict(named_args)

    def __str__(self) -> str:
        args = [str(a) for a in self.positional_args] + [
            f"{k}={v}" for k, v in self.named_args.items()
        ]
        return f"{self.name}({', '.join(args)})"

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, {!r})".format(
            self.__class__.__qualname__,
            self.name,
            self.positional_args,
            self.named_args,
        )

    def __eq__(self, o: object) -> bool:
        return (
            super().__eq__(o)
            and self.name == cast(__class__, o).name
            and self.positional_args == cast(__class__, o).positional_args
            and self.named_args == cast(__class__, o).named_args
        )


class BinaryOp(Expression):
    """
    The invocation of a binary operator.

    To avoid any precedence error in the generated code, operands that are also
    BinaryOps are always surrounded by braces (even when not necessary, as in
    "1 + (2 + 3)", as a more subtle behavior would increase the complexity of
    the implementation without much benefit.
    """

    def __init__(self, lhs: Expression, op: str, rhs: Expression) -> None:
        """
        :param lhs: Left-hand side operand of this operation.
        :param op: Name of the operator, like ``+``.
        :param rhs: Right-hand side operand of this operation.
        """
        super().__init__()
        self.lhs = lhs
        self.op = op
        self.rhs = rhs

    def __str__(self) -> str:
        operands = [self.lhs, self.rhs]
        return f" {self.op} ".join(
            f"({x})" if isinstance(x, BinaryOp) else str(x) for x in operands
        )

    def __eq__(self, o: object) -> bool:
        return (
            super().__eq__(o)
            and self.lhs == cast(__class__, o).lhs
            and self.op == cast(__class__, o).op
            and self.rhs == cast(__class__, o).rhs
        )


class Assignment(Statement):
    """
    The assignment of a value to a variable.

    For our purposes, we don't treat multiple assignment via tuples differently.
    We also don't support chained assignments such as ``a = b = 1``.
    """

    def __init__(self, lhs: str, rhs: Expression, comments: Sequence[str] = ()) -> None:
        """
        :param lhs: Variable name (or names) the *rhs* is assigned to.
        :param rhs: Expression which value is assigned to *lhs*.
        """
        super().__init__(comments)
        self.lhs = lhs
        self.rhs = rhs

    def lines(self, indent_level: int = 0, comments: bool = True) -> List[Line]:
        line = Line(f"{self.lhs} = {self.rhs}", indent_level)
        if comments:
            return self.attach_comment(line)
        return [line]

    def __eq__(self, o: object) -> bool:
        return (
            super().__eq__(o)
            and self.lhs == cast(__class__, o).lhs
            and self.rhs == cast(__class__, o).rhs
        )

    def __repr__(self) -> str:
        return "{}(lhs={!r}, rhs={!r}, comments={!r})".format(
            self.__class__.__qualname__, self.lhs, self.rhs, self.comments
        )


class IfElse(Statement):
    """
    The if/elif/else construct, where elif and else are optional and elif can
    be repeated.
    """

    def __init__(
        self,
        condition_blocks: Sequence[Tuple[Expression, Sequence[Statement]]],
        else_block: Optional[Sequence[Statement]] = None,
        comments: Sequence[str] = (),
    ) -> None:
        """
        :param condition_blocks: Pairs of condition and statements.
            Each pair is composed of an expression representing a condition,
            and a list of statements corresponding to that condition.
            This represents an if/elif/.../elif sequence, where there is always
            an "if" clause and an arbitrary number of "elif" clauses.
        :param else_block: Statements representing the "else" clause, if any.
        :raise ValueError: If there is not at least one element in
            *condition_blocks*.
        """
        super().__init__(comments)
        self.condition_blocks = [
            (cond, list(stmts)) for cond, stmts in condition_blocks
        ]
        self._assert_consistency()
        self.else_block = else_block

    def _assert_consistency(self):
        if not self.condition_blocks:
            raise ValueError("can't have an if without at least one block")

    def lines(self, indent_level: int = 0, comments: bool = True) -> List[Line]:
        self._assert_consistency()
        lines = []
        for i, block in enumerate(self.condition_blocks):
            keyword = "if" if i == 0 else "elif"
            lines.append(Line(f"{keyword} {block[0]}:", indent_level))
            lines.extend(
                [
                    line
                    for stmt in block[1]
                    for line in stmt.lines(indent_level + 1, comments)
                ]
                or [Line("pass", indent_level + 1)]
            )
        if self.else_block:
            lines.append(Line("else:", indent_level))
            lines.extend(
                [
                    line
                    for stmt in self.else_block
                    for line in stmt.lines(indent_level + 1, comments)
                ]
            )
        if comments:
            # There is always a first line, or _assert_consistency would fail.
            return [*self.attach_comment(lines[0]), *lines[1:]]
        return lines

    def __eq__(self, o: object) -> bool:
        return (
            super().__eq__(o)
            and self.condition_blocks == cast(__class__, o).condition_blocks
            and self.else_block == cast(__class__, o).else_block
        )

    def __repr__(self) -> str:
        return "{}(condition_blocks={!r}, else_block={!r}, comments={!r})".format(
            self.__class__.__qualname__,
            self.condition_blocks,
            self.else_block,
            self.comments,
        )


class Import(Statement):
    """
    The import statement in all its forms: ``import X``, ``import X as A``,
    ``from M import X``, ``from M import X as A``, and ``from M import X, Y``.

    Combined imports like ``from M import X, Y`` are split for simplicity.
    """

    def __init__(
        self,
        targets: Sequence[str],
        source: Optional[str] = None,
        alias: Optional[str] = None,
        comments: Sequence[str] = (),
    ) -> None:
        """
        :param targets: What is imported: *X* in :samp:`import {X}` and
            :samp:`from M import {X}`.
        :param source: From where *targets* are imported, if applicable: *M* in
            :samp:`from {M} import X`.
        :param alias: Alias for a single-element *targets*: *A* in
            :samp:`import X as {A}` and :samp:`from M import X as {A}`.
        :raise ValueError: If *targets* is empty, or if *alias* is specified
            even though there are multiple *targets*.
        """
        super().__init__(comments)
        self.targets = list(targets)
        self.source = source
        self.alias = alias
        self._assert_consistency()

    def _assert_consistency(self):
        if not self.targets:
            raise ValueError("expected at least one import target")
        if len(self.targets) > 1 and self.alias:
            raise ValueError("alias forbidden for multiple import targets")

    def lines(self, indent_level: int = 0, comments: bool = True) -> List[Line]:
        self._assert_consistency()
        import_kw = f"from {self.source} import" if self.source else "import"
        alias_clause = f" as {self.alias}" if self.alias else ""
        lines = [
            Line(f"{import_kw} {target}{alias_clause}", indent_level)
            for target in self.targets
        ]
        if comments:
            return [*self.comment_lines(indent_level), *lines]
        return lines

    def __eq__(self, o: object) -> bool:
        return (
            super().__eq__(o)
            and self.targets == cast(__class__, o).targets
            and self.source == cast(__class__, o).source
            and self.alias == cast(__class__, o).alias
        )

    def __repr__(self) -> str:
        return "{}(targets={!r}, source={!r}, alias={!r}, comments={!r})".format(
            self.__class__.__qualname__,
            self.targets,
            self.source,
            self.alias,
            self.comments,
        )


class Return(Statement):
    """The return statement."""

    def __init__(self, value: Expression, comments: Sequence[str] = ()) -> None:
        super().__init__(comments)
        self.value = value

    def lines(self, indent_level: int = 0, comments: bool = True) -> List[Line]:
        line = Line(f"return {self.value}", indent_level)
        if comments:
            return self.attach_comment(line)
        return [line]

    def __eq__(self, o: object) -> bool:
        return super().__eq__(o) and self.value == cast(self.__class__, o).value

    def __repr__(self) -> str:
        return "{}(value={!r}, comments={!r})".format(
            self.__class__.__qualname__, self.value, self.comments
        )


_T = TypeVar("_T")


@dataclass
class ExpressionView(Expression):
    """
    A "proxy" for an object that is not an :class:`Expression`.

    .. |Expr| replace:: :class:`Expression`
    .. |EV| replace:: :class:`ExpressionView`

    |EV| allows to mix non-|Expr| objects in the syntax tree, along with
    a function capable of transforming these objects into actual |Expr| objects
    at any time.
    This is useful when these objects are easier to manipulate than their |Expr|
    equivalent.

    .. |Request| replace:: :class:`Request <transformer.request.Request>`

    For example: any |Request| object can be converted into an equivalent |Expr|,
    but |Request| has a simpler API than |Expr| for request-oriented operations
    like accessing the URL, etc.
    |EV| can "wrap" a |Request| to pretend that the |Request| is an |Expr|
    (with all associated benefits of being part of the syntax tree), but still
    support the |Request| API.

    .. attribute:: target

        :any:`() → T <typing.Callable>` --
        A function (without parameters) returning the wrapped, non-|Expr| object.

        The benefit of :attr:`target` being a function (instead of a direct
        reference to the wrapped object) is that it allows to specify some
        mutable field of an object.
        See for example :class:`Task2 <transformer.task.Task2>`, which contains
        an |EV| wrapping its own :attr:`request <transformer.task.Task2.request>`
        attribute.
        If the value of that attribute changes, the |EV| will refer to the new
        value (found by accessing the attribute via *self*),
        not the old value (which would still be referenced by a non-callable
        :attr:`target`).

    .. attribute:: converter

        :any:`T → <typing.Callable>` |Expr| --
        A function capable of transforming the result of :attr:`target` into
        an |Expr|.
        The result of :attr:`converter` is computed each time this |EV| has to
        behave like an |Expr|, for example when passed as argument to :any:`str`.

    .. attribute:: name

        :any:`str` --
        Purely descriptive: makes the inspection of data structures containing
        |EV| objects more comfortable.
    """

    target: Callable[[], _T]
    converter: Callable[[_T], Expression]
    name: str

    def __str__(self) -> str:
        return str(self.converter(self.target()))



import enum
import warnings
from typing import Sequence, List, Union, Iterator

import transformer.plugins as plug
import transformer.python as py
from transformer.plugins.contracts import Plugin
from transformer.scenario import Scenario
from transformer.task import Task, Task2
from ._version import __version__

LOCUST_MAX_WAIT_DELAY = 10

LOCUST_MIN_WAIT_DELAY = 0

LOCUSTFILE_COMMENT = f"""
File automatically generated by Transformer v{__version__}:
https://github.com/zalando-incubator/Transformer
""".strip()


def _locust_task(task: Union[Task, Task2]) -> py.Function:
    """
    Transforms a Task into the Python code expected by Locust.

    This function is private because it does not return a complete Locust task
    (the @task decorator is missing) and should therefore not be used for that
    purpose by unsuspecting users.
    """
    if isinstance(task, Task):
        # TODO: remove when Task2 has replaced Task.
        #   See https://github.com/zalando-incubator/Transformer/issues/11.
        task = Task2.from_task(task)

    return py.Function(name=task.name, params=["self"], statements=task.statements)


class TaskSetType(enum.Enum):
    Set = "TaskSet"
    Sequence = "TaskSequence"


def locust_taskset(scenario: Scenario) -> py.Class:
    """
    Transforms a scenario (potentially containing other scenarios) into a Locust
    TaskSet definition.
    """
    if any(isinstance(child, (Task, Task2)) for child in scenario.children):
        ts_type = TaskSetType.Sequence
    else:
        ts_type = TaskSetType.Set

    fields: List[py.Statement] = []
    for i, child in enumerate(scenario.children, start=1):
        seq_decorator = f"seq_task({i})"
        if isinstance(child, (Task2, Task)):
            fields.append(py.Decoration(seq_decorator, _locust_task(child)))
        elif isinstance(child, Scenario):
            field = py.Decoration(f"task({child.weight})", locust_taskset(child))
            if ts_type is TaskSetType.Sequence:
                field = py.Decoration(seq_decorator, field)
            fields.append(field)
        else:
            wrong_type = child.__class__.__qualname__
            scenario_type = scenario.__class__.__qualname__
            raise TypeError(
                f"unexpected type {wrong_type} in {scenario_type}.children: {child!r}"
            )
    return py.Class(scenario.name, superclasses=[str(ts_type.value)], statements=fields)


def locust_classes(scenarios: Sequence[Scenario]) -> List[py.Class]:
    """
    Transforms scenarios into all Python classes needed by Locust (TaskSet and
    Locust classes).

    The only missing parts before a fully functional locustfile are:
    - integrating all necessary set-up/tear-down statements:
        - Python imports,
        - apply global plugins,
        - etc.
    - serializing everything via transformer.python.
    """
    classes = []
    for scenario in scenarios:
        taskset = locust_taskset(scenario)
        is_post_1 = py.BinaryOp(py.Symbol("LOCUST_MAJOR_VERSION"), ">=", py.Literal(1))
        tasks = py.IfElse(
            [
                (
                    is_post_1,
                    [py.Assignment("tasks", py.Literal([py.Symbol(taskset.name)]))],
                )
            ],
            [py.Assignment("task_set", py.Symbol(taskset.name))],
        )
        locust_class = py.Class(
            name=f"LocustFor{taskset.name}",
            superclasses=["HttpLocust"],
            statements=[
                tasks,
                py.Assignment("weight", py.Literal(scenario.weight)),
                py.Assignment("min_wait", py.Literal(LOCUST_MIN_WAIT_DELAY)),
                py.Assignment("max_wait", py.Literal(LOCUST_MAX_WAIT_DELAY)),
            ],
        )
        classes.append(taskset)
        classes.append(locust_class)
    return classes


def locust_detected_version() -> py.Program:
    return [
        py.Import(["LooseVersion"], source="distutils.version"),
        py.Import(["__version__"], source="locust"),
        py.OpaqueBlock("LOCUST_MAJOR_VERSION = LooseVersion(__version__).version[0]"),
    ]


def locust_imports() -> py.Program:
    is_post_1 = py.BinaryOp(
        py.Symbol("LOCUST_MAJOR_VERSION"),
        ">=",
        py.Literal(1),
    )
    imports_pre_1 = [
        py.Import(
            ["HttpLocust", "TaskSequence", "TaskSet", "seq_task", "task"],
            source="locust",
        )
    ]
    imports_post_1 = [
        py.Import(
            ["HttpUser", "SequentialTaskSet", "TaskSet", "task"],
            source="locust",
        ),
        py.Assignment("HttpLocust", py.Symbol("HttpUser")),
        py.Assignment("TaskSequence", py.Symbol("SequentialTaskSet")),
        py.Function("seq_task", ["_"], [py.Return(py.Symbol("task"))]),
    ]
    return [
        py.IfElse([(is_post_1, imports_post_1)], imports_pre_1),
    ]


def locust_program(scenarios: Sequence[Scenario]) -> py.Program:
    """
    Converts a ScenarioGroup into a Locust File.
    """
    global_code_blocks = {
        # TODO: Replace me with a plugin framework that accesses the full tree.
        #   See https://github.com/zalando-incubator/Transformer/issues/11.
        block_name: py.OpaqueBlock("\n".join(block), comments=[block_name])
        for scenario in scenarios
        for block_name, block in scenario.global_code_blocks.items()
    }

    return [
        py.Import(["re"], comments=[LOCUSTFILE_COMMENT]),
        *locust_detected_version(),
        *locust_imports(),
        *locust_classes(scenarios),
        *global_code_blocks.values(),
    ]


def locustfile_lines(
    scenarios: Sequence[Scenario], program_plugins: Sequence[Plugin]
) -> Iterator[str]:
    """
    Converts the provided scenarios into a stream of Python statements
    and iterate on the resulting lines.
    """
    program = plug.apply(program_plugins, locust_program(scenarios))
    for stmt in program:
        for line in stmt.lines():
            yield str(line)


def locustfile(scenarios: Sequence[Scenario]) -> str:
    """
    Simple wrapper around locustfile_lines joining all lines with "\n".

    This function is deprecated and will be removed in a future version.
    Do not rely on it.
    Reason: It does not provide significant value over locustfile_lines and has
    a less clear name and a less flexible API. It does not support new
    generation plugins contracts like OnPythonProgram.
    Deprecated since: v1.0.2.
    """
    warnings.warn(DeprecationWarning("locustfile: use locustfile_lines instead"))
    return "\n".join(locustfile_lines(scenarios, ()))



"""
:mod:`transformer.scenario` -- Grouping related tasks into scenarios
====================================================================

.. |Scenario| replace:: :class:`Scenario`

Transformer creates a |Scenario| object for each HAR file it reads, so each
:term:`task` (representing an HTTP request in a HAR file) is part of a
:term:`scenario`.
Transformer also creates |Scenario| objects for directories that contain
HAR files, so a :term:`scenario` can also be part of another :term:`scenario`.

To sum up, |Scenario| objects form a tree_, the leaves of which are all
:term:`task` objects.
This hierarchical structure maps cleanly to Locust's :class:`~locust.core.TaskSet`
objects, which can also be nested and have a weight.

.. _tree: https://en.wikipedia.org/wiki/Tree_(data_structure)
"""

import json
import logging
from collections import defaultdict
from pathlib import Path
from typing import Sequence, Mapping, Union, Set, List, Optional, Dict, Tuple

import dataclasses
from dataclasses import dataclass

import transformer.plugins as plug
from transformer.denylist import Denylist, get_empty
from transformer.naming import to_identifier
from transformer.plugins.contracts import Plugin
from transformer.request import Request
from transformer.task import Task, Task2

WEIGHT_FILE_SUFFIX = ".weight"
DEFAULT_WEIGHT = 1


class SkippableScenarioError(ValueError):  # noqa: B903
    """
    Raised when a Scenario cannot be created from the provided input path.

    If related to the creation of a Scenario B inside a larger Scenario A (i.e.
    B would be in A.children), A catches this exception, logs a warning, and
    moves on to the next potential child.
    """

    def __init__(self, scenario_path: Path, reason: Union[Exception, str]) -> None:
        self.path = scenario_path
        self.reason = reason


class DanglingWeightError(SkippableScenarioError):
    """
    Raised when a scenario directory contains weight files that don't correspond
    to any scenario.
    """

    pass


class CollidingScenariosError(SkippableScenarioError):
    """
    Raised when scenarios created from different paths end up having the same
    name.

    The only way this happens is if the paths are identical save for their
    extension (e.g. ``.har`` vs ``.json``), or if there is a bug (collision) in
    :func:`~transformer.naming.to_identifier` (which should never happen).
    """

    pass


class WeightValueError(ValueError):  # noqa: B903
    """
    Raised when the weight file associated to a scenario contains errors.
    """

    def __init__(self, scenario_path: Path, reason: Union[Exception, str]) -> None:
        self.path = scenario_path
        self.reason = reason


@dataclass
class Scenario:
    """
    A web browsing session that we want to emulate, i.e. a sequence of
    :term:`tasks <task>` to be performed.

    .. attribute:: name

        :any:`str` -- Name of the corresponding :class:`~locust.core.TaskSet`.

    .. attribute:: children

        |Sequence| [ :class:`~transformer.task.Task2` :data:`| <typing.Union>`
        :class:`Scenario` ] --
        Tasks and scenarios that are part of this scenario.

    .. attribute:: origin

        :data:`~typing.Optional` :class:`~pathlib.Path` --
        Path to the HAR file or directory this scenario represents.

    .. attribute:: weight
       :annotation: = 1

       :any:`int` -- Weight of this scenario.
       See :ref:`Specifying-weights` and :ref:`Hierarchical-scenarios` for details.
    """

    name: str
    children: Sequence[Union[Task, Task2, "Scenario"]]
    origin: Optional[Path]
    weight: int = 1

    @classmethod
    def from_path(
        cls,
        path: Path,
        plugins: Sequence[Plugin] = (),
        ts_plugins: Sequence[Plugin] = (),
        short_name: bool = False,
        denylist: Optional[Denylist] = None,
    ) -> "Scenario":
        """
        Makes a :class:`Scenario` (possibly containing sub-scenarios) out of
        the provided *path*, which may point to either:

        - a HAR file (like :file:`x/y/z.har`),

        - a scenario directory (a directory containing HAR files or other
          scenario directories).

        :raise SkippableScenarioError: if path is neither a directory nor a HAR file,
            or is a directory containing dangling weight files
        :param path: path to the HAR file or scenario directory.
        :param plugins: list of :term:`OnScenario` plugins to apply.
        :param ts_plugins: deprecated -- for backward compatibility only.
        :param short_name: whether the returned scenarios have names based only
            on their path's basename, instead of the full path.
            By default *False* to avoid generating homonym scenarios
            (and therefore homonym :class:~locust.core.TaskSet` classes),
            but *True* when generating sub-scenarios (:attr:`children`) from
            a directory *path* (because then the names are "scoped" by
            the parent directory).
        :param denylist: a set of urls to be excluded
        """
        if denylist is None:
            denylist = get_empty()

        if path.is_dir():
            return cls.from_dir(
                path,
                plugins=plugins,
                ts_plugins=ts_plugins,
                short_name=short_name,
                denylist=denylist,
            )
        else:
            return cls.from_har_file(
                path,
                plugins,
                ts_plugins=ts_plugins,
                short_name=short_name,
                denylist=denylist,
            )

    @classmethod
    def from_dir(
        cls,
        path: Path,
        plugins: Sequence[Plugin],
        ts_plugins: Sequence[Plugin],
        short_name: bool,
        denylist: Denylist,
    ) -> "Scenario":
        """
        Makes a :class:`Scenario` out of the provided directory *path*.

        *path* must represent a "scenario directory", which contains at least
        one HAR file or another scenario directory.
        Symbolic link loops are not checked but forbidden!

        There may exist a weight file :file:`{path}.weight`.
        If so, its contents will be used as :attr:`weight` after calling
        :meth:`weight_from_path`.

        Errors are handled this way:

        #. If *path* itself cannot be transformed into a scenario,
           raise :exc:`SkippableScenarioError`.

        #. For each child of *path*, apply (1) but catch the exception and
           display a warning about skipping that child.
           (If all children are skipped, (1) applies to *path* itself.)

        Therefore:

        - If the directory contains weight files that don't match any HAR file or
          subdirectory, an error is emitted as this is probably a mistake.

        - If the directory contains files or directories that cannot be converted
          into scenarios (e.g. non-JSON files or :file:`.git` directories),
          a message is emitted and the file or subdirectory is skipped.

        :param path: path to the directory.
        :param plugins: list of :term:`OnScenario` plugins to apply.
        :param ts_plugins: deprecated -- for backward compatibility only.
        :param short_name: whether to simplify the resulting
            :class:`~locust.core.TaskSet` class name. If *short_name* is *False*,
            that class name is guaranteed to be unique across all TaskSets of the
            locustfile, but this is generally not necessary and results in less
            readable class names.
        :param denylist: a sequence of urls to be excluded
        :raise SkippableScenarioError: if the directory contains dangling weight
            files or no sub-scenarios.
        """
        try:
            children = list(path.iterdir())
        except OSError as err:
            raise SkippableScenarioError(path, err)

        weight_files: Set[Path] = {
            child for child in children if child.suffix == WEIGHT_FILE_SUFFIX
        }

        scenarios: List[Scenario] = []
        for child in children:
            if child in weight_files:
                continue
            try:
                scenario = cls.from_path(
                    child,
                    plugins,
                    ts_plugins=ts_plugins,
                    short_name=True,
                    denylist=denylist,
                )
            except SkippableScenarioError as err:
                logging.warning(
                    "while searching for HAR files, skipping %s: %s", child, err.reason
                )
            else:
                scenarios.append(scenario)

        cls._check_dangling_weights(path, scenarios, weight_files)
        if not scenarios:
            raise SkippableScenarioError(path, "no scenarios inside the directory")
        cls._check_name_collisions(path, scenarios)

        return Scenario(
            name=to_identifier(path.with_suffix("").name if short_name else str(path)),
            children=tuple(scenarios),
            origin=path,
            weight=cls.weight_from_path(path),
        )

    @classmethod
    def _check_name_collisions(cls, path: Path, scenarios: List["Scenario"]):
        scenarios_by_name: Dict[str, List[Scenario]] = defaultdict(list)
        for s in scenarios:
            scenarios_by_name[s.name].append(s)
        colliding_paths: Set[Tuple[Path, ...]] = {
            tuple(x.origin for x in xs)
            for xs in scenarios_by_name.values()
            if len(xs) > 1
        }
        if colliding_paths:
            groups = "; ".join(
                " vs ".join(repr(s.name) for s in group) for group in colliding_paths
            )
            logging.error(
                "%s contains scenarios with colliding names: %s", path, groups
            )
            raise CollidingScenariosError(path, "scenarios have colliding names")

    @classmethod
    def _check_dangling_weights(cls, path, scenarios, weight_files):
        scenario_names = {s.origin.with_suffix("").name for s in scenarios}
        dangling_weight_files = [
            f for f in weight_files if f.with_suffix("").name not in scenario_names
        ]
        if dangling_weight_files:
            hint = ", ".join(str(f) for f in dangling_weight_files)
            logging.error(
                "%s contains weight files that don't correspond to any scenarios: %s",
                path,
                hint,
            )
            logging.info(
                "For any value of X, if there exists a weight file X.weight, "
                "there must exist either an X.har file or an X scenario subdirectory."
            )
            raise DanglingWeightError(path, "contains dangling weight files")

    @classmethod
    def from_har_file(
        cls,
        path: Path,
        plugins: Sequence[Plugin],
        ts_plugins: Sequence[Plugin],
        short_name: bool,
        denylist: Denylist,
    ) -> "Scenario":
        """
        Creates a Scenario given a HAR file.

        :raise SkippableScenarioError: if path is unreadable or not a HAR file
        """
        try:
            with path.open(encoding="utf-8") as file:
                har = json.load(file)
            requests = Request.all_from_har(har)
            tasks = Task.from_requests(requests, denylist)

            # TODO: Remove this when Contract.OnTaskSequence is removed.
            tasks = plug.apply(ts_plugins, tasks)

            # TODO: Remove Task-to-Task2 conversion once both are merged.
            tasks = tuple(plug.apply(plugins, Task2.from_task(t)) for t in tasks)

            return Scenario(
                name=to_identifier(
                    path.with_suffix("").name if short_name else str(path)
                ),
                children=tuple(tasks),
                origin=path,
                weight=cls.weight_from_path(path),
            )
        except (OSError, json.JSONDecodeError, UnicodeDecodeError) as err:
            raise SkippableScenarioError(path, err)

    @classmethod
    def weight_from_path(cls, path: Path) -> int:
        """
        Reads the weight file corresponding to path, or returns a default weight
        if the weight file doesn't exist.

        :param path: represents either a HAR file or a scenario directory
        :raise WeightValueError: if the weight file exists but its contents cannot be
            interpreted as a weight
        """
        weight_path = path.with_suffix(WEIGHT_FILE_SUFFIX)

        try:
            weight = weight_path.read_text().strip()
        except OSError as err:
            logging.info(
                f"No {weight_path} provided for {path}: "
                f"assigning default weight {DEFAULT_WEIGHT} ({err})"
            )
            return DEFAULT_WEIGHT

        if not weight.isdecimal() or int(weight) == 0:
            logging.error(
                "invalid weight file %s: weights must be positive integers, got %r",
                weight_path,
                weight,
            )
            raise WeightValueError(path, weight)

        return int(weight)

    @property
    def global_code_blocks(self) -> Mapping[str, Sequence[str]]:
        """
        .. deprecated:: 1.0.2

            This attribute is only kept for backward compatibility purposes.
            It exists because Transformer's first plugin system didn't have
            :term:`OnPythonProgram`, so plugins had to specify the top-level
            locustfile code blocks they needed (e.g. imports, global variables)
            at the :class:`Task` level and let the plugin system percolate these
            code blocks through the scenario tree.
            This explains why tasks have the similar
            :any:`transformer.task.Task2.global_code_blocks` field.
        """
        # TODO: Replace me with a plugin framework that accesses the full tree.
        #   See https://github.com/zalando-incubator/Transformer/issues/11.
        return {
            block_name: block_lines
            for child in self.children
            for block_name, block_lines in child.global_code_blocks.items()
        }

    def apply_plugins(self, plugins: Sequence[Plugin]) -> "Scenario":
        """
        Recursively builds a new scenario tree from the leaves by applying all
        *plugins* to each cloned scenario subtree.

        Does not do anything if *plugins* is empty.

        :param plugins: the plugins to apply.
            See :ref:`Specifying-plugins` for details.
        """
        if not plugins:
            return self

        children = [
            c.apply_plugins(plugins) if isinstance(c, Scenario) else c
            for c in self.children
        ]
        return plug.apply(plugins, dataclasses.replace(self, children=children))



"""
:mod:`transformer.request` -- HTTP requests read from HAR
=========================================================

Representation of HAR Request objects.
"""

import enum
from datetime import datetime
from types import MappingProxyType
from typing import Iterator, List, Optional
from urllib.parse import urlparse, SplitResult
from requests.structures import CaseInsensitiveDict

import pendulum
from dataclasses import dataclass

from transformer.naming import to_identifier


class HttpMethod(enum.Enum):
    """
    Enumeration of supported HTTP method types.
    """

    GET = enum.auto()  #: GET
    POST = enum.auto()  #: POST
    PUT = enum.auto()  #: PUT
    OPTIONS = enum.auto()  #: OPTIONS
    DELETE = enum.auto()  #: DELETE
    PATCH = enum.auto()  #: PATCH


@dataclass(frozen=True)
class QueryPair:
    """
    A pair of query parameters, as recorded in a HAR file (queryString__).

    __ http://www.softwareishard.com/blog/har-12-spec/#queryString
    """

    name: str
    value: str


@dataclass
class Request:
    """
    An HTTP request, as recorded in a HAR file (request__).

    __ http://www.softwareishard.com/blog/har-12-spec/#request

    Note that *post_data*, if present, will be a dict of the same format
    as recorded in the HAR file
    (postData__ -- although it is not consistently followed by HAR generators).

    __ http://www.softwareishard.com/blog/har-12-spec/#postData

    .. attribute:: timestamp

        :class:`~datetime.datetime` --
        Time at which the request was recorded.

    .. attribute:: method

        :class:`HttpMethod` --
        HTTP method of the request.

    .. attribute:: url

        :class:`urllib.parse.SplitResult` --
        URL targeted by the request.

    .. attribute:: har_entry

        :any:`dict` --
        A single record from entries `as recorded in a HAR file`__
        corresponding to the request, provided for read-only access.

        __ http://www.softwareishard.com/blog/har-12-spec/#entries

    .. attribute:: headers

        :any:`requests.structures.CaseInsensitiveDict` (reference__) --
        HTTP headers sent with the request.

        __ https://github.com/kennethreitz/requests/blob\
        /7e297ed95bdbd1018657f5d6000379ecdfa54423/requests/structures.py#L13

    .. attribute:: post_data
        :annotation: = None

        :data:`~typing.Optional` :any:`dict` --
        If :attr:`method` is ``POST``, the corresponding data payload.

    .. attribute:: query
        :annotation: = []

        :class:`~typing.List` of :class:`QueryPair` --
        Key-value arguments sent as part of the :attr:`url`'s `query string`__.

        __ https://en.wikipedia.org/wiki/Query_string

    .. attribute:: name
        :annotation: = None

        :data:`~typing.Optional` :any:`str` --
        Value provided for :class:`locust.clients.HttpSession`'s "dynamic"
        ``name`` parameter.
        See `Grouping requests to URLs with dynamic parameters`__ for details.

        __ https://docs.locust.io/en/stable/writing-a-locustfile.html#grouping-\
requests-to-urls-with-dynamic-parameters

    """

    timestamp: datetime
    method: HttpMethod
    url: SplitResult
    har_entry: dict
    headers: CaseInsensitiveDict = MappingProxyType({})
    post_data: Optional[dict] = None
    query: List[QueryPair] = ()
    name: Optional[str] = None

    def __post_init__(self):
        self.query = list(self.query)

    @classmethod
    def from_har_entry(cls, entry: dict) -> "Request":
        """
        Creates a request from a HAR entry__.

        __ http://www.softwareishard.com/blog/har-12-spec/#entries

        :raise KeyError: if *entry* is not a valid HAR "entry" object.
        :raise ValueError: if the ``request.startedDateTime`` value cannot be
            interpreted as a timestamp.
        """

        request = entry["request"]
        return Request(
            timestamp=pendulum.parse(entry["startedDateTime"]),
            method=HttpMethod[request["method"]],
            url=urlparse(request["url"]),
            har_entry=entry,
            name=None,
            headers=CaseInsensitiveDict(
                {d["name"]: d["value"] for d in request.get("headers", [])}
            ),
            post_data=request.get("postData"),
            query=[
                QueryPair(name=d["name"], value=d["value"])
                for d in request.get("queryString", [])
            ],
        )

    @classmethod
    def all_from_har(cls, har: dict) -> Iterator["Request"]:
        """
        Generates requests for all entries__ in a given HAR top-level object.

        __ http://www.softwareishard.com/blog/har-12-spec/#entries
        """

        for entry in har["log"]["entries"]:
            yield cls.from_har_entry(entry)

    def task_name(self) -> str:
        """
        Generates a simple name to be used as
        :attr:`~transformer.task.Task2.name` by the :term:`task` of this request.
        """

        return "_".join(
            (
                self.method.name,
                self.url.scheme,
                to_identifier(self.url.hostname),
                to_identifier(self.url.path),
                str(abs(hash(self))),
            )
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.timestamp,
                self.method,
                self.url,
                tuple(self.headers),
                repr(self.post_data) if self.post_data else None,
                tuple(self.query),
            )
        )



"""
:mod:`transformer.task` -- HTTP requests and related processing
===============================================================

Each HTTP request from a HAR file is seen by Transformer as a separate
:term:`task` to be eventually converted into a :any:`locust.core.task` function:

.. figure:: _static/basic-conversion.*
   :align: center

   *Transformer converts HAR requests into Locust tasks.*

:class:`~transformer.request.Request` only represents an HTTP request, not the
potential pre- and post-processing that could be desired in the same Locust task
(e.g. before or after the ``requests.get`` call).
Transformer's :term:`task` is an object encapsulating a
:class:`~transformer.request.Request` *and* that additional processing code,
in a one-to-one relationship with :any:`locust.core.task`-decorated functions.

However, Transformer's tasks have no notion of :ref:`weight <specifying-weights>`
or :ref:`grouping <hierarchical-scenarios>`: these come with
:term:`scenarios <scenario>`.
"""
import json
from collections import OrderedDict
from json import JSONDecodeError
from types import MappingProxyType
from typing import (
    Iterable,
    NamedTuple,
    Iterator,
    Sequence,
    Optional,
    Mapping,
    Dict,
    List,
    Tuple,
    cast,
)

import dataclasses
from dataclasses import dataclass
from requests.structures import CaseInsensitiveDict

import transformer.python as py
from transformer.denylist import on_denylist, Denylist, get_empty
from transformer.request import HttpMethod, Request, QueryPair

IMMUTABLE_EMPTY_DICT = MappingProxyType({})
TIMEOUT = 30
ACTION_INDENTATION_LEVEL = 12
JSON_MIME_TYPE = "application/json"


class LocustRequest(NamedTuple):
    """
    All parameters for the request performed by the Locust client object.

    .. deprecated:: 1.0.2
        Only used by :class:`Task`, which is itself deprecated.
        Use :class:`Task2` instead of :class:`Task`.
    """

    method: HttpMethod
    url: str
    headers: CaseInsensitiveDict = MappingProxyType({})
    post_data: dict = MappingProxyType({})
    query: Sequence[QueryPair] = ()
    name: Optional[str] = None

    @classmethod
    def from_request(cls, r: Request) -> "LocustRequest":
        return LocustRequest(
            method=r.method,
            url=repr(r.url.geturl()),
            headers=r.headers,
            post_data=r.post_data,
            query=r.query,
            name=repr(r.name or r.url.geturl()),
        )


@dataclass
class Task2:
    """
    Represents a :term:`task`, i.e. an HTTP request along with some optional
    pre- and post-processing code.

    .. attribute:: name

        :any:`str` --
        Name of the corresponding :any:`locust.core.task` function in
        the locustfile.

    .. attribute:: request

        :any:`transformer.request.Request` --
        HTTP request executed by this task.

    .. attribute:: statements

        :any:`Sequence <typing.Sequence>` of |Statement| --
        Body of the corresponding :any:`locust.core.task` function in the
        locustfile.

        One of these statements contains an |ExpressionView| pointing to
        :attr:`request`.
        The other statements (if any) represent pre- or post-processing code
        for that request, depending on whether they appear before or after the
        statement containing the |ExpressionView|.

        .. warning::

            Plugins should be careful if they replace the |ExpressionView|
            object found in :attr:`statements`.
            Other plugins should still be able to change :attr:`request` and
            expect to see these changes reflected in :attr:`statements` via
            |ExpressionView|.

    .. attribute:: global_code_blocks

        :any:`Mapping <typing.Mapping>` of
        :any:`str` to |Statement|

        .. deprecated:: 1.0.2

            This attribute is only kept for backward compatibility purposes.
            It exists because Transformer's first plugin system didn't have
            :term:`OnPythonProgram`, so plugins had to specify the top-level
            locustfile code blocks they needed (e.g. imports, global variables)
            at the :class:`Task` level and let the plugin system percolate these
            code blocks through the scenario tree.
            This explains why scenarios have the similar
            :any:`transformer.scenario.Scenario.global_code_blocks` field.
    """

    name: str
    request: Request
    statements: Sequence[py.Statement] = ()
    # TODO: Replace me with a plugin framework that accesses the full tree.
    #   See https://github.com/zalando-incubator/Transformer/issues/11.
    global_code_blocks: Mapping[str, Sequence[str]] = IMMUTABLE_EMPTY_DICT

    def __post_init__(self) -> None:
        self.statements = list(self.statements)
        self.global_code_blocks = {
            k: list(v) for k, v in self.global_code_blocks.items()
        }

    @classmethod
    def from_requests(cls, requests: Iterable[Request]) -> Iterator["Task2"]:
        """
        Generates a set of tasks from a given set of HTTP requests.

        Each request will be turned into an unevaluated function call
        (:class:`transformer.python.FunctionCall`)
        making the actual request.

        The returned tasks are ordered by increasing :any:`timestamp
        <transformer.request.Request.timestamp>` of the corresponding request.
        """
        # TODO: Update me when merging Task with Task2: "statements" needs to
        #   contain a ExpressionView to Task2.request.
        #   See what is done in from_task (but without the LocustRequest part).
        #   See https://github.com/zalando-incubator/Transformer/issues/11.
        for req in sorted(requests, key=lambda r: r.timestamp):
            if not on_denylist(req.url.geturl()):
                yield cls(name=req.task_name(), request=req, statements=...)

    @classmethod
    def from_task(cls, task: "Task") -> "Task2":
        # TODO: Remove me as soon as the old Task is no longer used and Task2 is
        #   renamed to Task.
        #   See https://github.com/zalando-incubator/Transformer/issues/11.
        t = cls(name=task.name, request=task.request)
        if task.locust_request:
            expr_view = py.ExpressionView(
                name="this task's request field",
                target=lambda: task.locust_request,
                converter=lreq_to_expr,
            )
        else:
            expr_view = py.ExpressionView(
                name="this task's request field",
                target=lambda: t.request,
                converter=req_to_expr,
            )
        t.statements = [
            *[py.OpaqueBlock(x) for x in task.locust_preprocessing],
            py.Assignment("response", expr_view),
            *[py.OpaqueBlock(x) for x in task.locust_postprocessing],
        ]
        return t


NOOP_HTTP_METHODS = {HttpMethod.GET, HttpMethod.OPTIONS, HttpMethod.DELETE}


def req_to_expr(r: Request) -> py.FunctionCall:
    url = py.Literal(str(r.url.geturl()))
    args: Dict[str, py.Expression] = OrderedDict(
        url=url,
        name=py.Literal(r.name) if r.name else url,
        timeout=py.Literal(TIMEOUT),
        allow_redirects=py.Literal(False),
    )
    if r.headers:
        args["headers"] = py.Literal(r.headers)

    if r.method is HttpMethod.POST:
        if r.post_data:
            rpd = RequestsPostData.from_har_post_data(r.post_data)
            args.update(rpd.as_kwargs())
    elif r.method in (HttpMethod.PUT, HttpMethod.PATCH):
        if r.post_data:
            rpd = RequestsPostData.from_har_post_data(r.post_data)
            args.update(rpd.as_kwargs())

        args.setdefault("params", py.Literal([]))
        cast(py.Literal, args["params"]).value.extend(
            _params_from_name_value_dicts([dataclasses.asdict(q) for q in r.query])
        )
    elif r.method not in NOOP_HTTP_METHODS:
        raise ValueError(f"unsupported HTTP method: {r.method!r}")

    method = r.method.name.lower()
    return py.FunctionCall(name=f"self.client.{method}", named_args=args)


def lreq_to_expr(lr: LocustRequest) -> py.FunctionCall:
    # TODO: Remove me once LocustRequest no longer exists.
    #   See https://github.com/zalando-incubator/Transformer/issues/11.
    url = _peel_off_repr(lr.url)
    name = _peel_off_repr(lr.name) if lr.name else url

    args: Dict[str, py.Expression] = OrderedDict(
        url=url,
        name=name,
        timeout=py.Literal(TIMEOUT),
        allow_redirects=py.Literal(False),
    )
    if lr.headers:
        args["headers"] = py.Literal(lr.headers)
    if lr.method is HttpMethod.POST:
        if lr.post_data:
            rpd = RequestsPostData.from_har_post_data(lr.post_data)
            args.update(rpd.as_kwargs())
    elif lr.method in (HttpMethod.PUT, HttpMethod.PATCH):
        if lr.post_data:
            rpd = RequestsPostData.from_har_post_data(lr.post_data)
            args.update(rpd.as_kwargs())

        args.setdefault("params", py.Literal([]))
        cast(py.Literal, args["params"]).value.extend(
            _params_from_name_value_dicts([dataclasses.asdict(q) for q in lr.query])
        )
    elif lr.method not in NOOP_HTTP_METHODS:
        raise ValueError(f"unsupported HTTP method: {lr.method!r}")

    method = lr.method.name.lower()
    return py.FunctionCall(name=f"self.client.{method}", named_args=args)


def _peel_off_repr(s: str) -> py.Literal:
    """
    Reverse the effect of LocustRequest's repr() calls on url and name.
    """
    if s.startswith("f"):
        return py.FString(eval(s[1:], {}, {}))
    return py.Literal(eval(s, {}, {}))


class Task(NamedTuple):
    """
    One step of "doing something" on a website.
    This basically represents a @task in Locust-speak.

    .. deprecated:: 1.0.2
        Use :class:`Task2` instead. :class:`Task` is kept for backward
        compatibility with existing plugins that have not yet migrated to
        :class:`Task2`.
        Transformer will automatically convert :class:`Task` objects into
        :class:`Task2` objects using :meth:`Task2.from_task`.
    """

    name: str
    request: Request
    locust_request: Optional[LocustRequest] = None
    locust_preprocessing: Sequence[str] = ()
    locust_postprocessing: Sequence[str] = ()
    global_code_blocks: Mapping[str, Sequence[str]] = MappingProxyType({})

    @classmethod
    def from_requests(
        cls, requests: Iterable[Request], denylist: Optional[Denylist] = None
    ) -> Iterator["Task"]:
        """
        Generates a set of Tasks from a given set of Requests.
        """
        if denylist is None:
            denylist = get_empty()

        for req in sorted(requests, key=lambda r: r.timestamp):
            if on_denylist(denylist, req.url.geturl()):
                continue
            else:
                yield cls(name=req.task_name(), request=req)

    def inject_headers(self, headers: dict):
        if self.locust_request is None:
            original_locust_request = LocustRequest.from_request(self.request)
        else:
            original_locust_request = self.locust_request

        new_locust_request = original_locust_request._replace(
            headers=CaseInsensitiveDict({**original_locust_request.headers, **headers})
        )
        task = self._replace(locust_request=new_locust_request)

        return task

    def replace_url(self, url: str):
        if self.locust_request is None:
            original_locust_request = LocustRequest.from_request(self.request)
        else:
            original_locust_request = self.locust_request

        new_locust_request = original_locust_request._replace(url=url)
        return self._replace(locust_request=new_locust_request)


@dataclass
class RequestsPostData:
    """
    Data to be sent via HTTP POST, along with which API of the requests library
    to use.
    """

    data: Optional[py.Literal] = None
    params: Optional[py.Literal] = None
    json: Optional[py.Literal] = None

    def as_kwargs(self) -> Dict[str, py.Expression]:
        return {k: v for k, v in dataclasses.asdict(self).items() if v is not None}

    @classmethod
    def from_har_post_data(cls, post_data: dict) -> "RequestsPostData":
        """
        Converts a HAR postData object into a RequestsPostData instance.

        :param post_data: a HAR "postData" object,
            see http://www.softwareishard.com/blog/har-12-spec/#postData.
        :raise ValueError: if *post_data* is invalid.
        """
        try:
            return _from_har_post_data(post_data)
        except ValueError as err:
            raise ValueError(f"invalid HAR postData object: {post_data!r}") from err


def _from_har_post_data(post_data: dict) -> RequestsPostData:
    mime_k = "mimeType"
    try:
        mime: str = post_data[mime_k]
    except KeyError:
        raise ValueError(f"missing {mime_k!r} field") from None

    rpd = RequestsPostData()

    # The "text" and "params" fields are supposed to be mutually
    # exclusive (according to the HAR spec) but nobody respects that.
    # Often, both text and params are provided for x-www-form-urlencoded.
    text_k, params_k = "text", "params"
    if text_k not in post_data and params_k not in post_data:
        raise ValueError(f"should contain {text_k!r} or {params_k!r}")

    _extract_text(mime, post_data, text_k, rpd)

    try:
        params = _params_from_post_data(params_k, post_data)
        if params is not None:
            rpd.params = py.Literal(params)
    except (KeyError, UnicodeEncodeError, TypeError) as err:
        raise ValueError("unreadable params field") from err

    return rpd


def _extract_text(
    mime: str, post_data: dict, text_k: str, rpd: RequestsPostData
) -> None:
    text = post_data.get(text_k)
    if mime == JSON_MIME_TYPE:
        if text is None:
            raise ValueError(f"missing {text_k!r} field for {JSON_MIME_TYPE} content")
        try:
            rpd.json = py.Literal(json.loads(text))
        except JSONDecodeError as err:
            raise ValueError(f"unreadable JSON from field {text_k!r}") from err
    elif text is not None:  # Probably application/x-www-form-urlencoded.
        try:
            rpd.data = py.Literal(text.encode())
        except UnicodeEncodeError as err:
            raise ValueError(f"cannot encode the {text_k!r} field in UTF-8") from err


def _params_from_post_data(
    key: str, post_data: dict
) -> Optional[List[Tuple[bytes, bytes]]]:
    """
    Extracts the *key* list from *post_data* and calls
    _params_from_name_value_dicts with that list.

    :raise TypeError: if the object at *key* is built using unexpected data types.
    """
    params = post_data.get(key)
    if params is None:
        return None
    if not isinstance(params, list):
        raise TypeError(f"the {key!r} field should be a list")
    return _params_from_name_value_dicts(params)


def _params_from_name_value_dicts(
    dicts: Iterable[Mapping[str, str]]
) -> List[Tuple[bytes, bytes]]:
    """
    Converts a HAR "params" element [0] into a list of tuples that can be used
    as value for requests' "params" keyword-argument.

    [0]: http://www.softwareishard.com/blog/har-12-spec/#params
    [1]: http://docs.python-requests.org/en/master/user/quickstart/
        #more-complicated-post-requests

    :raise KeyError: if one of the elements doesn't contain a "name" or "value" field.
    :raise UnicodeEncodeError: if an element's "name" or "value" string cannot
        be encoded in UTF-8.
    """
    return [(d["name"].encode(), d["value"].encode()) for d in dicts]



import warnings
import logging
import os
from typing import Set
import re

Denylist = Set[str]


def get_empty() -> Denylist:
    return set()


def from_file() -> Denylist:
    if os.path.exists(f"{os.getcwd()}/.urlignore"):
        warnings.warn(
            "Legacy .urlignore file detected - it will not be used! Rename it to '.ignore' if you want to use it, and read about the difference in the documentation."
        )
    denylist_file = f"{os.getcwd()}/.ignore"
    try:
        with open(denylist_file, encoding="utf-8") as file:
            return set(filter(None, [line.rstrip() for line in file]))
    except OSError as err:
        logging.debug("Could not read denylist file %s. Reason: %s", denylist_file, err)
        return get_empty()


def on_denylist(denylist: Denylist, url: str) -> bool:
    """
    Checks for matching URLs in an ignore file (denylist)
    from user's current directory.
    """
    for denylist_item in denylist:
        if re.search(denylist_item, url):
            return True
    return False



"""
:mod:`transformer` -- Main API
============================================

This module exports the functions that should cover most use-cases of any
Transformer user.
"""
from .transform import dumps, dump
from ._version import __version__ as package_version

__version__ = package_version

__all__ = ["dumps", "dump"]



"""
:mod:`transformer.plugins` -- Plugin System
===========================================

This module exposes the API needed to create your own Transformer plugins.
"""
from .resolve import resolve
from .contracts import plugin, Contract, apply, group_by_contract

__all__ = ["resolve", "plugin", "Contract", "apply", "group_by_contract"]



import logging
from typing import cast

from transformer.plugins import plugin, Contract
from transformer.request import Request
from transformer.scenario import Scenario
from transformer.task import Task


@plugin(Contract.OnScenario)
def f(s: Scenario) -> Scenario:
    first_req = first(s)
    logging.info(f"The first request was {first_req.url.geturl()}")
    return s


def first(s: Scenario) -> Request:
    while isinstance(s, Scenario):
        s = s.children[0]
    return cast(Task, s).request



"""
:mod:transformer.transform -- Entrypoint functions
==================================================

Defines user-facing functions performing all or most of the HAR-to-locustfile
conversion.
"""
import warnings
from pathlib import Path
from typing import Sequence, Union, Iterable, TextIO, Iterator, TypeVar

import transformer.plugins as plug
from transformer.locust import locustfile, locustfile_lines
from transformer.plugins import sanitize_headers, Contract
from transformer.plugins.contracts import Plugin
from transformer.scenario import Scenario
from transformer import denylist

DEFAULT_PLUGINS = (sanitize_headers.plugin,)


def transform(
    scenarios_path: Union[str, Path],
    plugins: Sequence[Plugin] = (),
    with_default_plugins: bool = True,
) -> str:
    """
    This function is deprecated and will be removed in a future version.
    Do not rely on it.
    Reason: It only accepts one scenario path at a time, and requires plugins
    to be already resolved (and therefore that users use
    transformer.plugins.resolve, which is kind of low-level). Both dumps & dump
    lift these constraints and have a more familiar naming
    (see json.dump/s, etc.).
    Deprecated since: v1.0.2.
    """
    warnings.warn(DeprecationWarning("transform: use dump or dumps instead"))
    if with_default_plugins:
        plugins = (*DEFAULT_PLUGINS, *plugins)
    return locustfile(
        [
            Scenario.from_path(
                Path(scenarios_path), plugins, denylist=denylist.from_file()
            )
        ]
    )


LaxPath = Union[str, Path]
PluginName = str


def dumps(
    scenario_paths: Iterable[LaxPath],
    plugins: Sequence[PluginName] = (),
    with_default_plugins: bool = True,
) -> str:
    """
    Transforms the provided *scenario_paths* using the provided *plugins*,
    and returns the resulting locustfile code as a string.

    See also: :func:`dump`

    :param scenario_paths: paths to scenario files (HAR) or directories
    :param plugins: names of plugins to use
    :param with_default_plugins: whether the default plugins should be used in
        addition to those provided (recommended: True)
    """
    return "\n".join(_dump_as_lines(scenario_paths, plugins, with_default_plugins))


def dump(
    file: TextIO,
    scenario_paths: Iterable[LaxPath],
    plugins: Sequence[PluginName] = (),
    with_default_plugins: bool = True,
) -> None:
    """
    Transforms the provided *scenario_paths* using the provided *plugins*,
    and writes the resulting locustfile code in the provided *file*.

    See also: :func:`dumps`

    :param file: an object with a `writelines` method (as specified by
        io.TextIOBase), e.g. `sys.stdout` or the result of `open`.
    :param scenario_paths: paths to scenario files (HAR) or directories.
    :param plugins: names of plugins to use.
    :param with_default_plugins: whether the default plugins should be used in
        addition to those provided (recommended: True).
    """
    file.writelines(
        intersperse("\n", _dump_as_lines(scenario_paths, plugins, with_default_plugins))
    )


def _dump_as_lines(
    scenario_paths: Iterable[LaxPath],
    plugins: Sequence[PluginName],
    with_default_plugins: bool,
) -> Iterator[str]:
    plugins = [p for name in plugins for p in plug.resolve(name)]
    if with_default_plugins:
        plugins = (*DEFAULT_PLUGINS, *plugins)

    plugins_for = plug.group_by_contract(plugins)
    scenarios = [
        Scenario.from_path(
            path,
            plugins_for[Contract.OnTask],
            plugins_for[Contract.OnTaskSequence],
            denylist=denylist.from_file(),
        ).apply_plugins(plugins_for[Contract.OnScenario])
        for path in scenario_paths
    ]

    yield from locustfile_lines(scenarios, plugins_for[Contract.OnPythonProgram])


T = TypeVar("T")


def intersperse(delim: T, iterable: Iterable[T]) -> Iterator[T]:
    """
    >>> list(intersperse(",", "a"))
    ['a']
    >>> list(intersperse(",", ""))
    []
    >>> list(intersperse(",", "abc"))
    ['a', ',', 'b', ',', 'c']
    >>> list(intersperse(",", ["a", "b", "c"]))
    ['a', ',', 'b', ',', 'c']
    """
    it = iter(iterable)
    try:
        yield next(it)
    except StopIteration:
        return
    for x in it:
        yield delim
        yield x



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['transformer', 'transformer.plugins']

package_data = \
{'': ['*']}

install_requires = \
['chevron>=0.13,<0.14',
 'docopt>=0.6.2,<0.7.0',
 'ecological>=1.6,<2.0',
 'pendulum>=2.0,<3.0',
 'requests>=2.21']

extras_require = \
{'docs': ['sphinx', 'sphinx-autodoc-typehints', 'sphinx-issues']}

entry_points = \
{'console_scripts': ['transformer = transformer.cli:script_entrypoint']}

setup_kwargs = {
    'name': 'har-transformer',
    'version': '3.0.0rc1',
    'description': 'A tool to convert HAR files into a locustfile.',
    'long_description': '\n.. image:: docs/_static/transformer.png\n   :alt: Transformer logo\n   :align: center\n\n|\n\n.. image:: https://travis-ci.org/zalando-incubator/Transformer.svg?branch=master\n   :alt: travis-ci status badge\n   :target: https://travis-ci.org/zalando-incubator/Transformer\n\n.. image:: https://badgen.net/pypi/v/har-transformer\n   :alt: pypi version badge\n   :target: https://pypi.org/project/har-transformer\n\n.. image:: https://api.codacy.com/project/badge/Grade/10b3feb4e4814429bf288b87443a6c72\n   :alt: code quality badge\n   :target: https://www.codacy.com/app/thilp/Transformer\n\n.. image:: https://api.codacy.com/project/badge/Coverage/10b3feb4e4814429bf288b87443a6c72\n   :alt: test coverage badge\n   :target: https://www.codacy.com/app/thilp/Transformer\n\n.. image:: https://badgen.net/badge/code%20style/black/000\n   :alt: Code style: Black\n   :target: https://github.com/ambv/black\n\n\nTransformer\n***********\n\nA **command-line tool** and **Python library** to convert web browser sessions\n(`HAR`_ files) into Locust_ load test scenarios ("locustfiles").\n\n.. _HAR: https://en.wikipedia.org/wiki/.har\n.. _Locust: https://locust.io/\n\nUse it to **replay HAR files** (storing recordings of interactions with your\nwebsite) **in load tests** with Locust_.\n\n.. contents::\n   :local:\n\nInstallation\n============\n\nInstall from PyPI::\n\n   pip install har-transformer\n\nInstall Locust to run your locustfiles::\n\n   pip install locust\n\nUsage\n=====\n\nExample HAR files are included in the ``examples/`` directory, try them\nout.\n\nCommand-line\n------------\n\n.. code:: bash\n\n   transformer my_har_files_directory/ >locustfile.py\n\nLibrary\n-------\n\n.. code:: python\n\n   import transformer\n\n   with open("locustfile.py", "w") as f:\n       transformer.dump(f, ["my_har_files_directory/"])\n\nDocumentation\n=============\n\nTake a look at our documentation_ for more details, including how to **generate\nHAR files**, **customize your scenarios**, use or write **plugins**, etc.\n\n.. _documentation: https://transformer.readthedocs.io/\n\nAuthors\n=======\n\n- **Serhii Cherniavskyi** — `@scherniavsky <https://github.com/scherniavsky>`_\n\n- **Thibaut Le Page** — `@thilp <https://github.com/thilp>`_\n\n- **Brian Maher** — `@bmaher <https://github.com/bmaher>`_\n\n- **Oliwia Zaremba** — `@tortila <https://github.com/tortila>`_\n\nSee also the list of contributors_ to this project.\n\n.. _contributors: https://transformer.readthedocs.io/en/latest/Contributors.html\n\nLicense\n=======\n\nThis project is licensed under the MIT license — see the LICENSE.md_ file for\ndetails.\n\n.. _LICENSE.md: https://github.com/zalando-incubator/Transformer/blob/master\n   /LICENSE.md\n',
    'author': 'Serhii Cherniavskyi',
    'author_email': 'serhii.cherniavskyi@zalando.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://transformer.readthedocs.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import importlib
import inspect
import logging
import sys
from os import getcwd
from types import ModuleType
from typing import Iterator

from transformer.plugins.contracts import (
    Plugin,
    Contract,
    InvalidContractError,
    contract,
    InvalidPluginError,
)


def resolve(name: str) -> Iterator[Plugin]:
    """
    Transform a plugin name into the corresponding, actual plugins.

    The name of a plugin is the name of a Python module containing (at least)
    one function decorated with @plugin (from the contracts module).
    The "resolve" function loads that module and returns these plugin functions
    found inside the module.

    :raise ImportError: if name does not match an accessible module.
    :raise TypeError: from load_load_plugins_from_module.
    :raise InvalidContractError: from load_load_plugins_from_module.
    :raise NoPluginError: from load_load_plugins_from_module.
    """
    sys.path.append(getcwd())
    module = importlib.import_module(name)

    yield from load_plugins_from_module(module)


class NoPluginError(ValueError):
    """
    Raised for Python modules that should but don't contain any plugin function.
    """


def load_plugins_from_module(module: ModuleType) -> Iterator[Plugin]:
    """
    :param module: Python module from which to load plugin functions.
    :raise TypeError: if module is not a Python module.
    :raise InvalidContractError: if a function is associated to an invalid contract.
    :raise NoPluginError: if module doesn't contain at least one plugin function.
    """
    if not inspect.ismodule(module):
        raise TypeError(f"expected a module, got {module!r}")

    nb_plugins = 0

    for _, obj in inspect.getmembers(module, inspect.isfunction):
        try:
            c = contract(obj)
        except InvalidPluginError:
            logging.debug(f"ignoring {_n(obj)}: not decorated with @plugin")
            continue

        if not isinstance(c, Contract):
            msg = f"{_n(obj)} associated to an invalid contract {c!r}"
            raise InvalidContractError(msg)

        nb_plugins += 1
        yield obj

    if nb_plugins < 1:
        raise NoPluginError(module)


def _n(x) -> str:
    return getattr(x, "__qualname__", None) or repr(x)
